@@ -44,6 +44,7 @@
 
 #include "graphene-euler.h"
 
+#include "graphene-alloc-private.h"
 #include "graphene-matrix.h"
 #include "graphene-quaternion.h"
 #include "graphene-vectors-private.h"
@@ -64,7 +65,7 @@
 graphene_euler_t *
 graphene_euler_alloc (void)
 {
-  return calloc (1, sizeof (graphene_euler_t));
+  return graphene_aligned_alloc0 (sizeof (graphene_euler_t), 1, 16);
 }
 
 /**
@@ -78,7 +79,33 @@ graphene_euler_alloc (void)
 void
 graphene_euler_free (graphene_euler_t *e)
 {
-  free (e);
+  graphene_aligned_free (e);
+}
+
+/*< private >
+ * graphene_euler_init_internal:
+ * @e: the #graphene_euler_t to initialize
+ * @x: rotation angle on the X axis, in radians
+ * @y: rotation angle on the Y axis, in radians
+ * @z: rotation angle on the Z axis, in radians
+ * @order: order of rotations
+ *
+ * Initializes a #graphene_euler_t using the given angles
+ * and order of rotation.
+ *
+ * Returns: (transfer none): the initialized #graphene_euler_t
+ */
+static graphene_euler_t *
+graphene_euler_init_internal (graphene_euler_t       *e,
+                              float                   rad_x,
+                              float                   rad_y,
+                              float                   rad_z,
+                              graphene_euler_order_t  order)
+{
+  graphene_vec3_init (&e->angles, rad_x, rad_y, rad_z);
+  e->order = order;
+
+  return e;
 }
 
 /**
@@ -102,7 +129,11 @@ graphene_euler_init (graphene_euler_t *e,
                      float             y,
                      float             z)
 {
-  return graphene_euler_init_with_order (e, x, y, z, GRAPHENE_EULER_ORDER_DEFAULT);
+  return graphene_euler_init_internal (e,
+                                       GRAPHENE_DEG_TO_RAD (x),
+                                       GRAPHENE_DEG_TO_RAD (y),
+                                       GRAPHENE_DEG_TO_RAD (z),
+                                       GRAPHENE_EULER_ORDER_DEFAULT);
 }
 
 /**
@@ -126,10 +157,11 @@ graphene_euler_init_with_order (graphene_euler_t       *e,
                                 float                   z,
                                 graphene_euler_order_t  order)
 {
-  graphene_vec3_init (&e->angles, x, y, z);
-  e->order = order;
-
-  return e;
+  return graphene_euler_init_internal (e,
+                                       GRAPHENE_DEG_TO_RAD (x),
+                                       GRAPHENE_DEG_TO_RAD (y),
+                                       GRAPHENE_DEG_TO_RAD (z),
+                                       order);
 }
 
 /**
@@ -265,10 +297,7 @@ graphene_euler_init_from_matrix (graphene_euler_t        *e,
       break;
     }
 
-  graphene_vec3_init (&e->angles,
-                      GRAPHENE_RAD_TO_DEG (x),
-                      GRAPHENE_RAD_TO_DEG (y),
-                      GRAPHENE_RAD_TO_DEG (z));
+  graphene_vec3_init (&e->angles, x, y, z);
 
   return e;
 }
@@ -349,10 +378,7 @@ graphene_euler_init_from_quaternion (graphene_euler_t            *e,
       break;
     }
 
-  graphene_vec3_init (&e->angles,
-                      GRAPHENE_RAD_TO_DEG (x),
-                      GRAPHENE_RAD_TO_DEG (y),
-                      GRAPHENE_RAD_TO_DEG (z));
+  graphene_vec3_init (&e->angles, x, y, z);
 
   return e;
 }
@@ -360,7 +386,8 @@ graphene_euler_init_from_quaternion (graphene_euler_t            *e,
 /**
  * graphene_euler_init_from_vec3:
  * @e: the #graphene_euler_t to initialize
- * @v: (nullable): a #graphene_vec3_t
+ * @v: (nullable): a #graphene_vec3_t containing the rotation
+ *   angles in degrees
  * @order: the order used to apply the rotations
  *
  * Initializes a #graphene_euler_t using the angles contained in a
@@ -379,7 +406,7 @@ graphene_euler_init_from_vec3 (graphene_euler_t       *e,
                                graphene_euler_order_t  order)
 {
   if (v != NULL)
-    graphene_vec3_init_from_vec3 (&e->angles, v);
+    graphene_vec3_scale (v, (GRAPHENE_PI / 180.f), &e->angles);
   else
     graphene_vec3_init_from_vec3 (&e->angles, graphene_vec3_zero ());
 
@@ -452,7 +479,7 @@ graphene_euler_equal (const graphene_euler_t *a,
 float
 graphene_euler_get_x (const graphene_euler_t *e)
 {
-  return graphene_vec3_get_x (&e->angles);
+  return GRAPHENE_RAD_TO_DEG (graphene_vec3_get_x (&e->angles));
 }
 
 /**
@@ -468,7 +495,7 @@ graphene_euler_get_x (const graphene_euler_t *e)
 float
 graphene_euler_get_y (const graphene_euler_t *e)
 {
-  return graphene_vec3_get_y (&e->angles);
+  return GRAPHENE_RAD_TO_DEG (graphene_vec3_get_y (&e->angles));
 }
 
 /**
@@ -484,7 +511,7 @@ graphene_euler_get_y (const graphene_euler_t *e)
 float
 graphene_euler_get_z (const graphene_euler_t *e)
 {
-  return graphene_vec3_get_z (&e->angles);
+  return GRAPHENE_RAD_TO_DEG (graphene_vec3_get_z (&e->angles));
 }
 
 /**
@@ -527,6 +554,7 @@ graphene_euler_to_vec3 (const graphene_euler_t *e,
                         graphene_vec3_t        *res)
 {
   graphene_vec3_init_from_vec3 (res, &e->angles);
+  graphene_vec3_scale (res, (180.f / GRAPHENE_PI), res);
 }
 
 /**
@@ -559,11 +587,9 @@ graphene_euler_to_matrix (const graphene_euler_t *e,
 {
   graphene_euler_order_t order = graphene_euler_get_order (e);
 
-  const float x = GRAPHENE_DEG_TO_RAD (graphene_vec3_get_x (&e->angles));
-  const float y = GRAPHENE_DEG_TO_RAD (graphene_vec3_get_y (&e->angles));
-  const float z = GRAPHENE_DEG_TO_RAD (graphene_vec3_get_z (&e->angles));
-
-  graphene_vec4_t row_x, row_y, row_z, row_w;
+  const float x = graphene_vec3_get_x (&e->angles);
+  const float y = graphene_vec3_get_y (&e->angles);
+  const float z = graphene_vec3_get_z (&e->angles);
 
   float c1, s1, c2, s2, c3, s3;
   float c3c2, s3c1, c3s2s1, s3s1;
@@ -595,9 +621,10 @@ graphene_euler_to_matrix (const graphene_euler_t *e,
          * ⎢ -s3 c3 0 ⎥ ⎢  0  1   0 ⎥ ⎢ 0  c1 s1 ⎥
          * ⎣   0  0 1 ⎦ ⎣ s2  0  c2 ⎦ ⎣ 0 -s1 c1 ⎦
          */
-        graphene_vec4_init (&row_x,  c3c2, s3c1 + c3s2s1, s3s1 - c3s2c1, 0.f);
-        graphene_vec4_init (&row_y, -s3c2, c3c1 - s3s2s1, c3s1 + s3s2c1, 0.f);
-        graphene_vec4_init (&row_z,    s2,         -c2s1,          c2c1, 0.f);
+        res->value.x = graphene_simd4f_init ( c3c2, s3c1 + c3s2s1, s3s1 - c3s2c1, 0.f);
+        res->value.y = graphene_simd4f_init (-s3c2, c3c1 - s3s2s1, c3s1 + s3s2c1, 0.f);
+        res->value.z = graphene_simd4f_init (   s2,         -c2s1,          c2c1, 0.f);
+        res->value.w = graphene_simd4f_init (  0.f,           0.f,           0.f, 1.f);
       }
       break;
 
@@ -607,9 +634,10 @@ graphene_euler_to_matrix (const graphene_euler_t *e,
          * ⎢ -s2 c3 0 ⎥ ⎢ 0  c2 s2 ⎥ ⎢  0 1   0 ⎥
          * ⎣   0  0 1 ⎦ ⎣ 0 -s2 c2 ⎦ ⎣ s1 0  c1 ⎦
          */
-        graphene_vec4_init (&row_x,  c3c1 + s3s2s1, s3c2, -c3s1 + s3s2c1, 0.f);
-        graphene_vec4_init (&row_y, -s3c1 + c3s2s1, c3c2,  s3s1 + c3s2c1, 0.f);
-        graphene_vec4_init (&row_z,           c2s1,  -s2,           c2c1, 0.f);
+        res->value.x = graphene_simd4f_init ( c3c1 + s3s2s1, s3c2, -c3s1 + s3s2c1, 0.f);
+        res->value.y = graphene_simd4f_init (-s3c1 + c3s2s1, c3c2,  s3s1 + c3s2c1, 0.f);
+        res->value.z = graphene_simd4f_init (          c2s1,  -s2,           c2c1, 0.f);
+        res->value.w = graphene_simd4f_init (           0.f,  0.f,            0.f, 1.f);
       }
       break;
 
@@ -619,9 +647,10 @@ graphene_euler_to_matrix (const graphene_euler_t *e,
          * ⎢ 0  c3 s3 ⎥ ⎢  0  1   0 ⎥ ⎢ -s1 c1 0 ⎥
          * ⎣ 0 -s3 c3 ⎦ ⎣ s2  0  c2 ⎦ ⎣   0  0 1 ⎦
          */
-        graphene_vec4_init (&row_x, c3c1 - s3s2s1, c3s1 + s3s2c1, -s3c2, 0.f);
-        graphene_vec4_init (&row_y,         -c2s1,          c2c1,    s2, 0.f);
-        graphene_vec4_init (&row_z, s3c1 + c3s2s1, s3s1 - c3s2c1,  c3c2, 0.f);
+        res->value.x = graphene_simd4f_init (c3c1 - s3s2s1, c3s1 + s3s2c1, -s3c2, 0.f);
+        res->value.y = graphene_simd4f_init (        -c2s1,          c2c1,    s2, 0.f);
+        res->value.z = graphene_simd4f_init (s3c1 + c3s2s1, s3s1 - c3s2c1,  c3c2, 0.f);
+        res->value.w = graphene_simd4f_init (          0.f,           0.f,   0.f, 1.f);
       }
       break;
 
@@ -631,9 +660,10 @@ graphene_euler_to_matrix (const graphene_euler_t *e,
          * ⎢ 0  c3 s3 ⎥ ⎢  0  1   0 ⎥ ⎢ -s1 c1 0 ⎥
          * ⎣ 0 -s3 c3 ⎦ ⎣ s2  0  c2 ⎦ ⎣   0  0 1 ⎦
          */
-        graphene_vec4_init (&row_x,          c2c1,          c2s1,  -s2, 0.f);
-        graphene_vec4_init (&row_y, s3s2c1 - c3s1, s3s2s1 + c3c1, s3c2, 0.f);
-        graphene_vec4_init (&row_z, c3s2c1 + s3s1, c3s2s1 - s3c1, c3c2, 0.f);
+        res->value.x = graphene_simd4f_init (         c2c1,          c2s1,  -s2, 0.f);
+        res->value.y = graphene_simd4f_init (s3s2c1 - c3s1, s3s2s1 + c3c1, s3c2, 0.f);
+        res->value.z = graphene_simd4f_init (c3s2c1 + s3s1, c3s2s1 - s3c1, c3c2, 0.f);
+        res->value.w = graphene_simd4f_init (          0.f,           0.f,  0.f, 1.f);
       }
       break;
 
@@ -643,9 +673,10 @@ graphene_euler_to_matrix (const graphene_euler_t *e,
          * ⎢ 0  c3 s3 ⎥ ⎢ -s2 c2 0 ⎥ ⎢  0 1   0 ⎥
          * ⎣ 0 -s3 c3 ⎦ ⎣   0  0 1 ⎦ ⎣ s1 0  c1 ⎦
          */
-        graphene_vec4_init (&row_x,           c2c1,    s2,          -c2s1, 0.f);
-        graphene_vec4_init (&row_y, -c3s2c1 + s3s1,  c3c2,  c3s2s1 + s3c1, 0.f);
-        graphene_vec4_init (&row_z,  s3s2c1 + c3s1, -s3c2, -s3s2s1 + c3c1, 0.f);
+        res->value.x = graphene_simd4f_init (          c2c1,    s2,          -c2s1, 0.f);
+        res->value.y = graphene_simd4f_init (-c3s2c1 + s3s1,  c3c2,  c3s2s1 + s3c1, 0.f);
+        res->value.z = graphene_simd4f_init ( s3s2c1 + c3s1, -s3c2, -s3s2s1 + c3c1, 0.f);
+        res->value.w = graphene_simd4f_init (           0.f,   0.f,            0.f, 1.f);
       }
       break;
 
@@ -655,22 +686,17 @@ graphene_euler_to_matrix (const graphene_euler_t *e,
          * ⎢  0 1   0 ⎥ ⎢ -s2 c2 0 ⎥ ⎢ 0  c1 s1 ⎥
          * ⎣ s3 0  c3 ⎦ ⎣   0  0 1 ⎦ ⎣ 0 -s1 c1 ⎦
          */
-        graphene_vec4_init (&row_x, c3c2, c3s2c1 + s3s1, c3s2s1 - s3c1, 0.f);
-        graphene_vec4_init (&row_y,  -s2,          c2c1,          c2s1, 0.f);
-        graphene_vec4_init (&row_z, s3c2, s3s2c1 - c3s1, s3s2s1 + c3c1, 0.f);
+        res->value.x = graphene_simd4f_init (c3c2, c3s2c1 + s3s1, c3s2s1 - s3c1, 0.f);
+        res->value.y = graphene_simd4f_init ( -s2,          c2c1,          c2s1, 0.f);
+        res->value.z = graphene_simd4f_init (s3c2, s3s2c1 - c3s1, s3s2s1 + c3c1, 0.f);
+        res->value.w = graphene_simd4f_init ( 0.f,           0.f,           0.f, 1.f);
       }
       break;
 
     default:
-      graphene_vec4_init (&row_x, 1.f, 0.f, 0.f, 0.f);
-      graphene_vec4_init (&row_y, 0.f, 1.f, 0.f, 0.f);
-      graphene_vec4_init (&row_z, 0.f, 0.f, 1.f, 0.f);
+      graphene_matrix_init_identity (res);
       break;
     }
-
-  graphene_vec4_init (&row_z, 0.f, 0.f, 0.f, 1.f);
-
-  graphene_matrix_init_from_vec4 (res, &row_x, &row_y, &row_z, &row_w);
 }
 
 /**
