@@ -32,7 +32,7 @@
  * The matrix is treated as row-major, i.e. it has four vectors (x, y, z, and
  * w) representing rows, and elements of each vector are a column:
  *
- * |[
+ * |[<!-- language="plain" -->
  *   | x |    | x.x   x.y   x.z   x.w |
  *   | y | -\ | y.x   y.y   y.z   y.w |
  *   | z | -/ | z.x   z.y   z.z   z.w |
@@ -58,6 +58,7 @@
 #include "graphene-point3d.h"
 #include "graphene-quad.h"
 #include "graphene-quaternion.h"
+#include "graphene-ray.h"
 #include "graphene-rect.h"
 #include "graphene-simd4x4f.h"
 #include "graphene-sphere.h"
@@ -113,6 +114,13 @@ graphene_matrix_to_float (const graphene_matrix_t *m,
   graphene_simd4x4f_to_float (&m->value, v);
 }
 
+static const float graphene_identity_matrix_floats[16] = {
+  1.f, 0.f, 0.f, 0.f,
+  0.f, 1.f, 0.f, 0.f,
+  0.f, 0.f, 1.f, 0.f,
+  0.f, 0.f, 0.f, 1.f,
+};
+
 /**
  * graphene_matrix_init_identity:
  * @m: a #graphene_matrix_t
@@ -126,7 +134,7 @@ graphene_matrix_to_float (const graphene_matrix_t *m,
 graphene_matrix_t *
 graphene_matrix_init_identity (graphene_matrix_t *m)
 {
-  graphene_simd4x4f_init_identity (&m->value);
+  graphene_simd4x4f_init_from_float (&m->value, graphene_identity_matrix_floats);
 
   return m;
 }
@@ -507,12 +515,13 @@ graphene_matrix_is_2d (const graphene_matrix_t *m)
 bool
 graphene_matrix_is_backface_visible (const graphene_matrix_t *m)
 {
-  graphene_matrix_t tmp;
+  graphene_simd4x4f_t tmp;
 
-  graphene_matrix_inverse (m, &tmp);
+  if (!graphene_simd4x4f_inverse (&m->value, &tmp))
+    return false;
 
   /* inverse.zz < 0 */
-  return graphene_matrix_get_value (&tmp, 2, 2) < 0.f;
+  return graphene_simd4f_get_z (tmp.z) < 0.f;
 }
 
 /**
@@ -528,7 +537,11 @@ graphene_matrix_is_backface_visible (const graphene_matrix_t *m)
 bool
 graphene_matrix_is_singular (const graphene_matrix_t *m)
 {
-  return graphene_matrix_determinant (m) == 0.0f;
+  graphene_simd4f_t det;
+
+  graphene_simd4x4f_determinant (&m->value, &det, NULL);
+
+  return fabsf (graphene_simd4f_get_x (det)) <= GRAPHENE_FLOAT_EPSILON;
 }
 
 /**
@@ -546,7 +559,7 @@ graphene_matrix_is_singular (const graphene_matrix_t *m)
  *
  * The arguments map to the following matrix layout:
  *
- * |[
+ * |[<!-- language="plain" -->
  *   | xx yx |   |  a  b  0 |
  *   | xy yy | = |  c  d  0 |
  *   | x0 y0 |   | tx ty  1 |
@@ -591,7 +604,7 @@ graphene_matrix_init_from_2d (graphene_matrix_t *m,
  *
  * The returned values have the following layout:
  *
- * |[
+ * |[<!-- language="plain" -->
  *   | xx yx |   |  a  b  0 |
  *   | xy yy | = |  c  d  0 |
  *   | x0 y0 |   | tx ty  1 |
@@ -691,7 +704,7 @@ graphene_matrix_get_row (const graphene_matrix_t *m,
 
 /**
  * graphene_matrix_get_value:
- * @m: a #grapheme_matrix_t
+ * @m: a #graphene_matrix_t
  * @row: the row index
  * @col: the column index
  *
@@ -884,18 +897,23 @@ graphene_matrix_transform_rect (const graphene_matrix_t *m,
                                 const graphene_rect_t   *r,
                                 graphene_quad_t         *res)
 {
-  graphene_point_t points[4];
   graphene_point_t ret[4];
 
-  graphene_rect_get_top_left (r, &points[0]);
-  graphene_rect_get_top_right (r, &points[1]);
-  graphene_rect_get_bottom_right (r, &points[2]);
-  graphene_rect_get_bottom_left (r, &points[3]);
+#define TRANSFORM_POINT(matrix,rect,corner,out_p)   do { \
+  graphene_simd4f_t __s; \
+  graphene_point_t __p; \
+  graphene_rect_get_ ## corner (rect, &__p); \
+  __s = graphene_simd4f_init (__p.x, __p.y, 0.f, 0.f); \
+  graphene_simd4x4f_vec3_mul (&matrix->value, &__s, &__s); \
+  out_p.x = graphene_simd4f_get (__s, 0); \
+  out_p.y = graphene_simd4f_get (__s, 1);           } while (0)
 
-  graphene_matrix_transform_point (m, &points[0], &ret[0]);
-  graphene_matrix_transform_point (m, &points[1], &ret[1]);
-  graphene_matrix_transform_point (m, &points[2], &ret[2]);
-  graphene_matrix_transform_point (m, &points[3], &ret[3]);
+  TRANSFORM_POINT (m, r, top_left, ret[0]);
+  TRANSFORM_POINT (m, r, top_right, ret[1]);
+  TRANSFORM_POINT (m, r, bottom_right, ret[2]);
+  TRANSFORM_POINT (m, r, bottom_left, ret[3]);
+
+#undef TRANSFORM_POINT
 
   graphene_quad_init (res, &ret[0], &ret[1], &ret[2], &ret[3]);
 }
@@ -917,33 +935,54 @@ graphene_matrix_transform_bounds (const graphene_matrix_t *m,
                                   const graphene_rect_t   *r,
                                   graphene_rect_t         *res)
 {
-  graphene_point_t points[4];
   graphene_point_t ret[4];
   float min_x, min_y;
   float max_x, max_y;
-  int i;
 
-  graphene_rect_get_top_left (r, &points[0]);
-  graphene_rect_get_top_right (r, &points[1]);
-  graphene_rect_get_bottom_left (r, &points[2]);
-  graphene_rect_get_bottom_right (r, &points[3]);
+#define TRANSFORM_POINT(matrix,rect,corner,out_p)   do { \
+  graphene_simd4f_t __s; \
+  graphene_point_t __p; \
+  graphene_rect_get_ ## corner (rect, &__p); \
+  __s = graphene_simd4f_init (__p.x, __p.y, 0.f, 0.f); \
+  graphene_simd4x4f_vec3_mul (&matrix->value, &__s, &__s); \
+  out_p.x = graphene_simd4f_get (__s, 0); \
+  out_p.y = graphene_simd4f_get (__s, 1);           } while (0)
 
-  graphene_matrix_transform_point (m, &points[0], &ret[0]);
-  graphene_matrix_transform_point (m, &points[1], &ret[1]);
-  graphene_matrix_transform_point (m, &points[2], &ret[2]);
-  graphene_matrix_transform_point (m, &points[3], &ret[3]);
+  TRANSFORM_POINT (m, r, top_left, ret[0]);
+  TRANSFORM_POINT (m, r, top_right, ret[1]);
+  TRANSFORM_POINT (m, r, bottom_right, ret[2]);
+  TRANSFORM_POINT (m, r, bottom_left, ret[3]);
 
-  min_x = max_x = ret[0].x;
-  min_y = max_y = ret[0].y;
+#undef TRANSFORM_POINT
 
-  for (i = 1; i < 4; i += 1)
-    {
-      min_x = MIN (ret[i].x, min_x);
-      min_y = MIN (ret[i].y, min_y);
+#if 0
+  {
+    int i;
 
-      max_x = MAX (ret[i].x, max_x);
-      max_y = MAX (ret[i].y, max_y);
-    }
+    min_x = max_x = ret[0].x;
+    min_y = max_y = ret[0].y;
+
+    for (i = 1; i < 4; i += 1)
+      {
+        min_x = MIN (ret[i].x, min_x);
+        min_y = MIN (ret[i].y, min_y);
+
+        max_x = MAX (ret[i].x, max_x);
+        max_y = MAX (ret[i].y, max_y);
+      }
+  }
+#else
+  {
+    const graphene_simd4f_t vx = graphene_simd4f_init (ret[0].x, ret[1].x, ret[2].x, ret[3].x);
+    const graphene_simd4f_t vy = graphene_simd4f_init (ret[0].y, ret[1].y, ret[2].y, ret[3].y);
+
+    min_x = graphene_simd4f_get_x (graphene_simd4f_min_val (vx));
+    min_y = graphene_simd4f_get_x (graphene_simd4f_min_val (vy));
+
+    max_x = graphene_simd4f_get_x (graphene_simd4f_max_val (vx));
+    max_y = graphene_simd4f_get_x (graphene_simd4f_max_val (vy));
+  }
+#endif
 
   graphene_rect_init (res, min_x, min_y, max_x - min_x, max_y - min_y);
 }
@@ -965,18 +1004,15 @@ graphene_matrix_transform_sphere (const graphene_matrix_t *m,
                                   const graphene_sphere_t *s,
                                   graphene_sphere_t       *res)
 {
-  graphene_point3d_t center;
   float max_scale;
 
-  graphene_point3d_init_from_vec3 (&center, &s->center);
-  graphene_matrix_transform_point3d (m, &center, &center);
+  graphene_simd4x4f_point3_mul (&m->value, &s->center.value, &res->center.value);
 
-  max_scale = graphene_simd4f_get_x (graphene_simd4f_dot3 (m->value.x, m->value.x));
-  max_scale = fmaxf (max_scale, graphene_simd4f_get_x (graphene_simd4f_dot3 (m->value.y, m->value.y)));
-  max_scale = fmaxf (max_scale, graphene_simd4f_get_x (graphene_simd4f_dot3 (m->value.z, m->value.z)));
-  max_scale = sqrtf (max_scale);
+  max_scale = graphene_simd4f_dot3_scalar (m->value.x, m->value.x);
+  max_scale = fmaxf (max_scale, graphene_simd4f_dot3_scalar (m->value.y, m->value.y));
+  max_scale = fmaxf (max_scale, graphene_simd4f_dot3_scalar (m->value.z, m->value.z));
 
-  graphene_sphere_init (res, &center, s->radius * max_scale);
+  res->radius = s->radius * sqrtf (max_scale);
 }
 
 /**
@@ -1007,6 +1043,34 @@ graphene_matrix_transform_box (const graphene_matrix_t *m,
   graphene_box_init_from_vectors (res, 8, points);
 }
 
+/**
+ * graphene_matrix_transform_ray:
+ * @m: a #graphene_matrix_t
+ * @r: a #graphene_ray_t
+ * @res: (out caller-allocates): return location for the
+ *   transformed ray
+ *
+ * Transform a #graphene_ray_t using the given matrix @m.
+ *
+ * Since: 1.4
+ */
+void
+graphene_matrix_transform_ray (const graphene_matrix_t *m,
+                               const graphene_ray_t    *r,
+                               graphene_ray_t          *res)
+{
+  graphene_vec3_t origin, direction;
+
+  graphene_vec3_add (&r->direction, &r->origin, &direction);
+  graphene_matrix_transform_vec3 (m, &direction, &direction);
+
+  graphene_matrix_transform_vec3 (m, &r->origin, &origin);
+
+  graphene_vec3_subtract (&direction, &origin, &direction);
+
+  graphene_ray_init_from_vec3 (res, &origin, &direction);
+}
+
 /**
  * graphene_matrix_project_point:
  * @m: a #graphene_matrix_t
@@ -1023,25 +1087,22 @@ graphene_matrix_project_point (const graphene_matrix_t *m,
                                const graphene_point_t  *p,
                                graphene_point_t        *res)
 {
-  graphene_vec3_t pa, qa;
-  graphene_vec3_t pback, qback, uback;
-  float p_z, u_z, t;
+  graphene_simd4f_t pa, pb, pc;
+  float a[3], b[3];
+  float t;
 
-  graphene_vec3_init (&pa, p->x, p->y, 0.0f);
-  graphene_vec3_init (&qa, p->x, p->y, 1.0f);
+  pa = graphene_simd4f_init (p->x, p->y, 0.f, 0.f);
+  pb = graphene_simd4f_init (p->x, p->y, 1.f, 0.f);
 
-  graphene_matrix_transform_vec3 (m, &pa, &pback);
-  graphene_matrix_transform_vec3 (m, &qa, &qback);
+  graphene_simd4x4f_vec3_mul (&m->value, &pa, &pa);
+  graphene_simd4x4f_vec3_mul (&m->value, &pb, &pb);
+  pc = graphene_simd4f_sub (pa, pb);
 
-  graphene_vec3_subtract (&qback, &pback, &uback);
+  graphene_simd4f_dup_3f (pa, a);
+  graphene_simd4f_dup_3f (pc, b);
+  t = -a[2] / b[2];
 
-  p_z = graphene_vec3_get_z (&pback);
-  u_z = graphene_vec3_get_z (&uback);
-  t = -p_z / u_z;
-
-  graphene_point_init (res,
-                       graphene_vec3_get_x (&pback) + t * graphene_vec3_get_x (&uback),
-                       graphene_vec3_get_y (&pback) + t * graphene_vec3_get_y (&uback));
+  graphene_point_init (res, a[0] + t * b[0], a[1] + t * b[1]);
 }
 
 /**
@@ -1153,7 +1214,9 @@ graphene_matrix_untransform_point (const graphene_matrix_t *m,
 
   if (graphene_matrix_is_2d (m))
     {
-      graphene_matrix_inverse (m, &inverse);
+      if (!graphene_matrix_inverse (m, &inverse))
+        return false;
+
       graphene_matrix_transform_point (&inverse, p, res);
       return true;
     }
@@ -1162,7 +1225,9 @@ graphene_matrix_untransform_point (const graphene_matrix_t *m,
   if (!graphene_rect_contains_point (&bounds_t, p))
     return false;
 
-  graphene_matrix_inverse (m, &inverse);
+  if (!graphene_matrix_inverse (m, &inverse))
+    return false;
+
   graphene_matrix_project_point (&inverse, p, res);
 
   return true;
@@ -1193,7 +1258,9 @@ graphene_matrix_untransform_bounds (const graphene_matrix_t *m,
 
   if (graphene_matrix_is_2d (m))
     {
-      graphene_matrix_inverse (m, &inverse);
+      if (!graphene_matrix_inverse (m, &inverse))
+        return;
+
       graphene_matrix_transform_bounds (&inverse, r, res);
       return;
     }
@@ -1205,14 +1272,17 @@ graphene_matrix_untransform_bounds (const graphene_matrix_t *m,
       return;
     }
 
-  graphene_matrix_inverse (m, &inverse);
+  if (!graphene_matrix_inverse (m, &inverse))
+    return;
+
   graphene_matrix_project_rect_bounds (&inverse, &rect, res);
 }
 
 /**
  * graphene_matrix_unproject_point3d:
  * @projection: a #graphene_matrix_t for the projection matrix
- * @modelview: a #graphene_matrix_t for the modelview matrix
+ * @modelview: a #graphene_matrix_t for the modelview matrix; this is
+ *   the inverse of the modelview used when projecting the point
  * @point: a #graphene_point3d_t with the coordinates of the point
  * @res: (out caller-allocates): return location for the unprojected
  *   point
@@ -1228,23 +1298,24 @@ graphene_matrix_unproject_point3d (const graphene_matrix_t  *projection,
                                    const graphene_point3d_t *point,
                                    graphene_point3d_t       *res)
 {
-  graphene_matrix_t inv_projection, tmp;
-  graphene_vec4_t v;
+  graphene_simd4x4f_t tmp;
+  graphene_simd4f_t v;
+  float values[4];
   float inv_w;
 
-  graphene_matrix_inverse (projection, &inv_projection);
-  graphene_matrix_multiply (&inv_projection, modelview, &tmp);
+  if (!graphene_simd4x4f_inverse (&projection->value, &tmp))
+    return;
 
-  graphene_vec4_init (&v, point->x, point->y, point->z, 1.f);
-  graphene_matrix_transform_vec4 (&tmp, &v, &v);
+  graphene_simd4x4f_matrix_mul (&tmp, &modelview->value, &tmp);
 
-  inv_w = 1.f / graphene_vec4_get_w (&v);
-  graphene_vec4_scale (&v, inv_w, &v);
+  v = graphene_simd4f_init (point->x, point->y, point->z, 1.f);
+  graphene_simd4x4f_vec4_mul (&tmp, &v, &v);
 
-  graphene_point3d_init (res,
-                         graphene_vec4_get_x (&v),
-                         graphene_vec4_get_y (&v),
-                         graphene_vec4_get_z (&v));
+  inv_w = 1.f / graphene_simd4f_get_w (v);
+  v = graphene_simd4f_mul (v, graphene_simd4f_splat (inv_w));
+
+  graphene_simd4f_dup_4f (v, values);
+  graphene_point3d_init (res, values[0], values[1], values[2]);
 }
 
 /**
@@ -1302,19 +1373,16 @@ graphene_matrix_rotate_euler (graphene_matrix_t      *m,
                               const graphene_euler_t *e)
 {
   graphene_quaternion_t q;
-  graphene_matrix_t rot;
 
   graphene_quaternion_init_from_euler (&q, e);
-  graphene_quaternion_to_matrix (&q, &rot);
-  graphene_matrix_multiply (m, &rot, m);
+  graphene_matrix_rotate_quaternion (m, &q);
 }
 
 static inline void
-graphene_matrix_rotate_internal (graphene_simd4x4f_t *m,
-                                 float                angle,
-                                 graphene_simd4f_t    axis)
+graphene_matrix_rotate_internal (graphene_simd4x4f_t     *m,
+                                 float                    rad,
+                                 const graphene_simd4f_t  axis)
 {
-  float rad = GRAPHENE_DEG_TO_RAD (angle);
   graphene_simd4x4f_t rot_m;
 
   graphene_simd4x4f_rotation (&rot_m, rad, axis);
@@ -1337,7 +1405,7 @@ graphene_matrix_rotate (graphene_matrix_t     *m,
                         float                  angle,
                         const graphene_vec3_t *axis)
 {
-  graphene_matrix_rotate_internal (&m->value, angle, axis->value);
+  graphene_matrix_rotate_internal (&m->value, GRAPHENE_DEG_TO_RAD (angle), axis->value);
 }
 
 /**
@@ -1354,7 +1422,8 @@ void
 graphene_matrix_rotate_x (graphene_matrix_t *m,
                           float              angle)
 {
-  graphene_matrix_rotate_internal (&m->value, angle, graphene_simd4f_init (1.f, 0.f, 0.f, 0.f));
+  graphene_matrix_rotate_internal (&m->value, GRAPHENE_DEG_TO_RAD (angle),
+                                   graphene_simd4f_init (1.f, 0.f, 0.f, 0.f));
 }
 
 /**
@@ -1371,7 +1440,8 @@ void
 graphene_matrix_rotate_y (graphene_matrix_t *m,
                           float              angle)
 {
-  graphene_matrix_rotate_internal (&m->value, angle, graphene_simd4f_init (0.f, 1.f, 0.f, 0.f));
+  graphene_matrix_rotate_internal (&m->value, GRAPHENE_DEG_TO_RAD (angle),
+                                   graphene_simd4f_init (0.f, 1.f, 0.f, 0.f));
 }
 
 /**
@@ -1388,7 +1458,8 @@ void
 graphene_matrix_rotate_z (graphene_matrix_t *m,
                           float              angle)
 {
-  graphene_matrix_rotate_internal (&m->value, angle, graphene_simd4f_init (0.f, 0.f, 1.f, 0.f));
+  graphene_matrix_rotate_internal (&m->value, GRAPHENE_DEG_TO_RAD (angle),
+                                   graphene_simd4f_init (0.f, 0.f, 1.f, 0.f));
 }
 
 /**
@@ -1478,17 +1549,17 @@ graphene_matrix_skew_yz (graphene_matrix_t *m,
   m->value.z = graphene_simd4f_madd (m_y, graphene_simd4f_splat (factor), m_z);
 }
 
-static void
+static inline void
 graphene_matrix_transpose_transform_vec4 (const graphene_matrix_t *m,
                                           const graphene_vec4_t   *v,
                                           graphene_vec4_t         *res)
 {
   float x, y, z, w;
 
-  x = graphene_simd4f_get_x (graphene_simd4f_sum (graphene_simd4f_mul (m->value.x, v->value)));
-  y = graphene_simd4f_get_x (graphene_simd4f_sum (graphene_simd4f_mul (m->value.y, v->value)));
-  z = graphene_simd4f_get_x (graphene_simd4f_sum (graphene_simd4f_mul (m->value.z, v->value)));
-  w = graphene_simd4f_get_x (graphene_simd4f_sum (graphene_simd4f_mul (m->value.w, v->value)));
+  x = graphene_simd4f_get_x (graphene_simd4f_dot4 (m->value.x, v->value));
+  y = graphene_simd4f_get_x (graphene_simd4f_dot4 (m->value.y, v->value));
+  z = graphene_simd4f_get_x (graphene_simd4f_dot4 (m->value.z, v->value));
+  w = graphene_simd4f_get_x (graphene_simd4f_dot4 (m->value.w, v->value));
 
   graphene_vec4_init (res, x, y, z, w);
 }
@@ -1518,13 +1589,15 @@ graphene_matrix_transpose (const graphene_matrix_t *m,
  *
  * Inverts the given matrix.
  *
+ * Returns: `true` if the matrix is invertible
+ *
  * Since: 1.0
  */
-void
+bool
 graphene_matrix_inverse (const graphene_matrix_t *m,
                          graphene_matrix_t       *res)
 {
-  graphene_simd4x4f_inverse (&m->value, &res->value);
+  return graphene_simd4x4f_inverse (&m->value, &res->value);
 }
 
 /**
@@ -1565,13 +1638,16 @@ graphene_matrix_normalize (const graphene_matrix_t *m,
   graphene_simd4f_t n;
   float ww;
 
-  ww = graphene_matrix_get_value (m, 3, 3);
-  n = graphene_simd4f_splat (ww);
+  ww = graphene_simd4f_get_w (m->value.w);
+  if (ww == 0.f)
+    return;
 
-  res->value.x = graphene_simd4f_div (m->value.x, n);
-  res->value.y = graphene_simd4f_div (m->value.y, n);
-  res->value.z = graphene_simd4f_div (m->value.z, n);
-  res->value.w = graphene_simd4f_div (m->value.w, n);
+  n = graphene_simd4f_splat (1.f / ww);
+
+  res->value.x = graphene_simd4f_mul (m->value.x, n);
+  res->value.y = graphene_simd4f_mul (m->value.y, n);
+  res->value.z = graphene_simd4f_mul (m->value.z, n);
+  res->value.w = graphene_simd4f_mul (m->value.w, n);
 }
 
 /**
@@ -1648,58 +1724,76 @@ graphene_matrix_get_z_scale (const graphene_matrix_t *m)
 #define XZ_SHEAR        1
 #define YZ_SHEAR        2
 
+#define M_11    0
+#define M_12    1
+#define M_21    2
+#define M_22    3
+
 static bool
 matrix_decompose_2d (const graphene_matrix_t *m,
-                     graphene_point3d_t      *scale_r,
-                     float                    shear_r[3],
-                     graphene_quaternion_t   *rotate_r,
-                     graphene_point3d_t      *translate_r)
+                     graphene_point_t        *translate_r,
+                     graphene_point_t        *scale_r,
+                     double                  *angle_r,
+                     float                    m_r[4])
 {
-  float A = graphene_matrix_get_value (m, 0, 0);
-  float B = graphene_matrix_get_value (m, 1, 0);
-  float C = graphene_matrix_get_value (m, 0, 1);
-  float D = graphene_matrix_get_value (m, 1, 1);
-  float scale_x, scale_y;
-  float shear_xy;
-  float rotate;
+  float row0x = graphene_matrix_get_value (m, 0, 0);
+  float row0y = graphene_matrix_get_value (m, 1, 0);
+  float row1x = graphene_matrix_get_value (m, 0, 1);
+  float row1y = graphene_matrix_get_value (m, 1, 1);
+  float angle;
+  float det;
 
-  if (A * D == B * C)
+  if (fabsf (row0x * row1y - row0y * row1x) < FLT_EPSILON)
     return false;
 
-  scale_x = sqrtf (A * A + B * B);
-  A /= scale_x;
-  B /= scale_x;
+  graphene_point_init (translate_r,
+                       graphene_matrix_get_value (m, 3, 0),
+                       graphene_matrix_get_value (m, 3, 1));
 
-  shear_xy = A * C + B * D;
-  C -= A * shear_xy;
-  D -= B * shear_xy;
+  scale_r->x = sqrtf (row0x * row0x + row0y * row0y);
+  scale_r->y = sqrtf (row1x * row1x + row1y * row1y);
 
-  scale_y = sqrtf (C * C + D * D);
-  C /= scale_y;
-  D /= scale_y;
-  shear_xy /= scale_y;
-
-  if (A * D < B * C)
+  det = row0x * row1y - row0y * row1x;
+  if (det < 0)
     {
-      A = -A;
-      B = -B;
-      C = -C;
-      D = -D;
-
-      shear_xy = -shear_xy;
-      scale_x = -scale_x;
+      if (row0x < row1y)
+        scale_r->x = -scale_r->x;
+      else
+        scale_r->y = -scale_r->y;
     }
 
-  rotate = atan2f (B, A);
-  graphene_quaternion_init (rotate_r, 0.f, 0.f, sin (rotate / 2.f), cos (rotate / 2.f));
+  if (scale_r->x != 0.f)
+    {
+      row0x = row0x * (1.f / scale_r->x);
+      row0y = row0y * (1.f / scale_r->y);
+    }
 
-  shear_r[XY_SHEAR] = shear_xy;
-  graphene_point3d_init (scale_r, scale_x, scale_y, 1.f);
+  if (scale_r->y != 0.f)
+    {
+      row1x = row1x * (1.f / scale_r->x);
+      row1y = row1y * (1.f / scale_r->y);
+    }
 
-  graphene_point3d_init (translate_r,
-                         graphene_matrix_get_value (m, 3, 0),
-                         graphene_matrix_get_value (m, 3, 1),
-                         0.f);
+  angle = atan2 (row0y, row0x);
+
+  if (angle)
+    {
+      double sn = -row0y, cs = row0x;
+      double m11 = row0x, m12 = row0y;
+      double m21 = row1x, m22 = row1y;
+
+      row0x = cs * m11 + sn * m21;
+      row0y = cs * m12 + sn * m22;
+      row1x = -sn * m11 + cs * m21;
+      row1y = -sn * m12 + cs * m22;
+    }
+
+  m_r[M_11] = row0x;
+  m_r[M_12] = row0y;
+  m_r[M_21] = row1x;
+  m_r[M_22] = row1y;
+
+  *angle_r = GRAPHENE_RAD_TO_DEG (angle);
 
   return true;
 }
@@ -1715,6 +1809,7 @@ matrix_decompose_3d (const graphene_matrix_t *m,
   graphene_matrix_t local, perspective;
   float shear_xy, shear_xz, shear_yz;
   float scale_x, scale_y, scale_z;
+  graphene_simd4f_t perspective_v;
   graphene_simd4f_t cross;
 
   if (graphene_matrix_get_value (m, 3, 3) == 0.f)
@@ -1735,22 +1830,32 @@ matrix_decompose_3d (const graphene_matrix_t *m,
   if (graphene_matrix_determinant (&perspective) == 0.f)
     return false;
 
+  perspective_v = graphene_simd4f_init (graphene_simd4f_get_w (local.value.x),
+                                        graphene_simd4f_get_w (local.value.y),
+                                        graphene_simd4f_get_w (local.value.z),
+                                        graphene_simd4f_get_w (local.value.w));
+
   /* isolate the perspective */
-  if (graphene_simd4f_is_zero3 (local.value.w))
+  if (!graphene_simd4f_is_zero3 (perspective_v))
     {
       graphene_matrix_t tmp;
 
       /* perspective_r is the right hand side of the equation */
-      perspective_r->value = local.value.w;
+      perspective_r->value = perspective_v;
 
       /* solve the equation by inverting perspective and multiplying
-       * the inverse with the perspective vector
+       * the inverse with the perspective vector; we don't need to
+       * check if the matrix is invertible here because we just checked
+       * whether the determinant is not zero.
        */
       graphene_matrix_inverse (&perspective, &tmp);
       graphene_matrix_transpose_transform_vec4 (&tmp, perspective_r, perspective_r);
 
-      /* clear the perspective partition */
-      local.value.w = graphene_simd4f_init (0.f, 0.f, 0.f, 1.f);
+      /* Clear the perspective component */
+      local.value.x = graphene_simd4f_merge_w (local.value.x, 0.f);
+      local.value.y = graphene_simd4f_merge_w (local.value.y, 0.f);
+      local.value.z = graphene_simd4f_merge_w (local.value.z, 0.f);
+      local.value.w = graphene_simd4f_merge_w (local.value.w, 1.f);
     }
   else
     graphene_vec4_init (perspective_r, 0.f, 0.f, 0.f, 1.f);
@@ -1797,7 +1902,7 @@ matrix_decompose_3d (const graphene_matrix_t *m,
    * coordinate system flip. if the determinant is -1, then
    * negate the matrix and the scaling factors
    */
-  cross = graphene_simd4f_dot4 (local.value.x, graphene_simd4f_cross3 (local.value.y, local.value.z));
+  cross = graphene_simd4f_dot3 (local.value.x, graphene_simd4f_cross3 (local.value.y, local.value.z));
   if (graphene_simd4f_get_x (cross) < 0.f)
     {
       scale_x *= -1.f;
@@ -1840,70 +1945,161 @@ graphene_matrix_interpolate (const graphene_matrix_t *a,
                              double                   factor,
                              graphene_matrix_t       *res)
 {
-  graphene_point3d_t scale_a = { 1.f, 1.f, 1.f }, translate_a;
-  graphene_quaternion_t rotate_a;
-  float shear_a[3] = { 0.f, 0.f, 0.f };
-
-  graphene_point3d_t scale_b = { 1.f, 1.f, 1.f }, translate_b;
-  graphene_quaternion_t rotate_b;
-  float shear_b[3] = { 0.f, 0.f, 0.f };
-
-  graphene_point3d_t scale_r = { 1.f, 1.f, 1.f }, translate_r;
-  graphene_quaternion_t rotate_r;
-  float shear;
-
   bool success = false;
 
+  graphene_matrix_init_identity (res);
+
+  /* Special case the decomposition if we're interpolating between two
+   * affine transformations.
+   */
   if (graphene_matrix_is_2d (a) &&
       graphene_matrix_is_2d (b))
     {
-      success |= matrix_decompose_2d (a, &scale_a, shear_a, &rotate_a, &translate_a);
-      success |= matrix_decompose_2d (b, &scale_b, shear_b, &rotate_b, &translate_b);
+      graphene_point_t translate_a = GRAPHENE_POINT_INIT (0.f, 0.f);
+      graphene_point_t translate_b = GRAPHENE_POINT_INIT (0.f, 0.f);
+      graphene_point_t translate_res;
+      graphene_point_t scale_a = GRAPHENE_POINT_INIT (1.f, 1.f);
+      graphene_point_t scale_b = GRAPHENE_POINT_INIT (1.f, 1.f);
+      graphene_point_t scale_res;
+      double rotate_a, rotate_b, rotate_res;
+      float m_a[4], m_b[4], m_res[4];
+      float rot_sin, rot_cos;
+      graphene_simd4x4f_t tmp_m;
 
+      success |= matrix_decompose_2d (a, &translate_a, &scale_a, &rotate_a, m_a);
+      success |= matrix_decompose_2d (b, &translate_b, &scale_b, &rotate_b, m_b);
+
+      /* If we cannot decompose either matrix we bail out with an identity */
+      if (!success)
+        return;
+
+      /* Flip the scaling factor and angle so they are consistent */
+      if ((scale_a.x < 0 && scale_b.y < 0) || (scale_a.y < 0 && scale_b.x < 0 ))
+        {
+          scale_a.x = -scale_a.x;
+          scale_a.y = -scale_a.y;
+
+          rotate_a += (rotate_a < 0) ? 180 : -180;
+        }
+
+      /* Do not rotate "the long way around" */
+      if (!rotate_a)
+        rotate_a = 360;
+      if (!rotate_b)
+        rotate_b = 360;
+
+      if (fabs (rotate_a - rotate_b) > 180)
+        {
+          if (rotate_a > rotate_b)
+            rotate_a -= 360;
+          else
+            rotate_b -= 360;
+        }
+
+      graphene_point_interpolate (&translate_a, &translate_b, factor, &translate_res);
+      graphene_point_interpolate (&scale_a, &scale_b, factor, &scale_res);
+      rotate_res = graphene_lerp (rotate_a, rotate_b, factor);
+
+      /* Interpolate each component of the (2,2) matrices */
+      {
+        graphene_simd4f_t tmp_va = graphene_simd4f_init_4f (m_a);
+        graphene_simd4f_t tmp_vb = graphene_simd4f_init_4f (m_b);
+        graphene_simd4f_t tmp_vres = graphene_simd4f_interpolate (tmp_va, tmp_vb, factor);
+
+        graphene_simd4f_dup_4f (tmp_vres, m_res);
+      }
+
+      /* Initialize using the transposed (2,2) matrix */
+      res->value.x = graphene_simd4f_init (m_res[M_11], m_res[M_21], 0.f, 0.f);
+      res->value.y = graphene_simd4f_init (m_res[M_12], m_res[M_22], 0.f, 0.f);
+      res->value.z = graphene_simd4f_init (        0.f,         0.f, 1.f, 0.f);
+
+      /* Translate */
+      res->value.w = graphene_simd4f_init (translate_res.x * m_res[M_11] + translate_res.y * m_res[M_21],
+                                           translate_res.x * m_res[M_12] + translate_res.y * m_res[M_22],
+                                           0.f,
+                                           1.f);
+
+      /* Rotate using a (2,2) rotation matrix */
+      graphene_sincos (GRAPHENE_DEG_TO_RAD (rotate_res), &rot_sin, &rot_cos);
+      tmp_m = graphene_simd4x4f_init (graphene_simd4f_init (rot_cos, -rot_sin, 0.f, 0.f),
+                                      graphene_simd4f_init (rot_sin,  rot_cos, 0.f, 0.f),
+                                      graphene_simd4f_init (    0.f,      0.f, 1.f, 0.f),
+                                      graphene_simd4f_init (    0.f,      0.f, 0.f, 1.f));
+      graphene_simd4x4f_matrix_mul (&res->value, &tmp_m, &res->value);
+
+      /* Scale */
+      graphene_simd4x4f_scale (&tmp_m, scale_res.x, scale_res.y, 1.f);
+      graphene_simd4x4f_matrix_mul (&res->value, &tmp_m, &res->value);
     }
   else
     {
+      graphene_point3d_t scale_a = { 1.f, 1.f, 1.f }, translate_a;
+      graphene_quaternion_t rotate_a;
+      float shear_a[3] = { 0.f, 0.f, 0.f };
+
+      graphene_point3d_t scale_b = { 1.f, 1.f, 1.f }, translate_b;
+      graphene_quaternion_t rotate_b;
+      float shear_b[3] = { 0.f, 0.f, 0.f };
+
+      graphene_point3d_t scale_r = { 1.f, 1.f, 1.f }, translate_r;
+      graphene_quaternion_t rotate_r;
+      float shear;
+
       graphene_vec4_t perspective_a;
       graphene_vec4_t perspective_b;
+      graphene_simd4f_t tmp;
 
       success |= matrix_decompose_3d (a, &scale_a, shear_a, &rotate_a, &translate_a, &perspective_a);
       success |= matrix_decompose_3d (b, &scale_b, shear_b, &rotate_b, &translate_b, &perspective_b);
 
-      res->value.w = graphene_simd4f_interpolate (perspective_a.value,
-                                                  perspective_b.value,
-                                                  factor);
+      /* If we cannot decompose either matrix we bail out with an identity */
+      if (!success)
+        return;
+
+      /* Interpolate the perspective row */
+      tmp = graphene_simd4f_interpolate (perspective_a.value, perspective_b.value, factor);
+      res->value.x = graphene_simd4f_init (1.f, 0.f, 0.f, graphene_simd4f_get_x (tmp));
+      res->value.y = graphene_simd4f_init (0.f, 1.f, 0.f, graphene_simd4f_get_y (tmp));
+      res->value.z = graphene_simd4f_init (0.f, 0.f, 1.f, graphene_simd4f_get_z (tmp));
+      res->value.w = graphene_simd4f_init (0.f, 0.f, 0.f, graphene_simd4f_get_w (tmp));
+
+      /* Translate */
+      graphene_point3d_interpolate (&translate_a, &translate_b, factor, &translate_r);
+      graphene_matrix_translate (res, &translate_r);
+
+      /* Rotate */
+      graphene_quaternion_slerp (&rotate_a, &rotate_b, factor, &rotate_r);
+      graphene_matrix_rotate_quaternion (res, &rotate_r);
+
+      /* Skew */
+      shear = graphene_lerp (shear_a[YZ_SHEAR], shear_b[YZ_SHEAR], factor);
+      if (shear != 0.f)
+        graphene_matrix_skew_yz (res, shear);
+
+      shear = graphene_lerp (shear_a[XZ_SHEAR], shear_b[XZ_SHEAR], factor);
+      if (shear != 0.f)
+        graphene_matrix_skew_xz (res, shear);
+
+      shear = graphene_lerp (shear_a[XY_SHEAR], shear_b[XY_SHEAR], factor);
+      if (shear != 0.f)
+        graphene_matrix_skew_xy (res, shear);
+
+      /* Scale */
+      graphene_point3d_interpolate (&scale_a, &scale_b, factor, &scale_r);
+      if (scale_r.x != 1.f && scale_r.y != 1.f && scale_r.z != 1.f)
+        graphene_matrix_scale (res, scale_r.x, scale_r.y, scale_r.z);
     }
-
-  /* If we cannot decompose either matrix we bail out with an identity */
-  if (!success)
-    {
-      graphene_matrix_init_identity (res);
-      return;
-    }
-
-  graphene_point3d_interpolate (&translate_a, &translate_b, factor, &translate_r);
-  graphene_matrix_translate (res, &translate_r);
-
-  graphene_quaternion_slerp (&rotate_a, &rotate_b, factor, &rotate_r);
-  graphene_matrix_rotate_quaternion (res, &rotate_r);
-
-  shear = graphene_lerp (shear_a[YZ_SHEAR], shear_b[YZ_SHEAR], factor);
-  if (shear != 0.f)
-    graphene_matrix_skew_yz (res, shear);
-
-  shear = graphene_lerp (shear_a[XZ_SHEAR], shear_b[XZ_SHEAR], factor);
-  if (shear != 0.f)
-    graphene_matrix_skew_xz (res, shear);
-
-  shear = graphene_lerp (shear_a[XY_SHEAR], shear_b[XY_SHEAR], factor);
-  if (shear != 0.f)
-    graphene_matrix_skew_xy (res, shear);
-
-  graphene_point3d_interpolate (&scale_a, &scale_b, factor, &scale_r);
-  if (scale_r.x != 1.f && scale_r.y != 1.f && scale_r.z != 1.f)
-    graphene_matrix_scale (res, scale_r.x, scale_r.y, scale_r.z);
 }
 
+#undef M_11
+#undef M_12
+#undef M_21
+#undef M_22
+#undef XY_SHEAR
+#undef XZ_SHEAR
+#undef YZ_SHEAR
+
 /**
  * graphene_matrix_print:
  * @m: The matrix to print
@@ -1920,7 +2116,7 @@ graphene_matrix_print (const graphene_matrix_t *m)
   for (i = 0; i < 4; i++)
     {
       fprintf (stderr,
-               "%.5f %.5f %.5f %.5f\n",
+               "| %+.6f %+.6f %+.6f %+.6f |\n",
                graphene_matrix_get_value (m, i, 0),
                graphene_matrix_get_value (m, i, 1),
                graphene_matrix_get_value (m, i, 2),
