@@ -246,23 +246,19 @@ void
 graphene_quaternion_to_matrix (const graphene_quaternion_t *q,
                                graphene_matrix_t           *m)
 {
-  graphene_simd4f_t m_x, m_y, m_z, m_w;
-
-  m_x = graphene_simd4f_init (1.f - 2.f * (q->y * q->y + q->z * q->z),
-                                    2.f * (q->x * q->y + q->w * q->z),
-                                    2.f * (q->x * q->z - q->w * q->y),
-                              0.f);
-  m_y = graphene_simd4f_init (      2.f * (q->x * q->y - q->w * q->z),
-                              1.f - 2.f * (q->x * q->x + q->z * q->z),
-                                    2.f * (q->y * q->z + q->w * q->x),
-                              0.f);
-  m_z = graphene_simd4f_init (      2.f * (q->x * q->z + q->w * q->y),
-                                    2.f * (q->y * q->z - q->w * q->x),
-                              1.f - 2.f * (q->x * q->x + q->y * q->y),
-                              0.f);
-  m_w = graphene_simd4f_init (0.f, 0.f, 0.f, 1.f);
-
-  m->value = graphene_simd4x4f_init (m_x, m_y, m_z, m_w);
+  m->value.x = graphene_simd4f_init (1.f - 2.f * (q->y * q->y + q->z * q->z),
+                                           2.f * (q->x * q->y + q->w * q->z),
+                                           2.f * (q->x * q->z - q->w * q->y),
+                                     0.f);
+  m->value.y = graphene_simd4f_init (      2.f * (q->x * q->y - q->w * q->z),
+                                     1.f - 2.f * (q->x * q->x + q->z * q->z),
+                                           2.f * (q->y * q->z + q->w * q->x),
+                                     0.f);
+  m->value.z = graphene_simd4f_init (      2.f * (q->x * q->z + q->w * q->y),
+                                           2.f * (q->y * q->z - q->w * q->x),
+                                     1.f - 2.f * (q->x * q->x + q->y * q->y),
+                                     0.f);
+  m->value.w = graphene_simd4f_init (0.f, 0.f, 0.f, 1.f);
 }
 
 /**
@@ -336,17 +332,41 @@ graphene_quaternion_init_from_angles (graphene_quaternion_t *q,
                                       float                  deg_y,
                                       float                  deg_z)
 {
-  float rad_x, rad_y, rad_z;
+  return graphene_quaternion_init_from_radians (q,
+                                                GRAPHENE_DEG_TO_RAD (deg_x),
+                                                GRAPHENE_DEG_TO_RAD (deg_y),
+                                                GRAPHENE_DEG_TO_RAD (deg_z));
+}
+
+/**
+ * graphene_quaternion_init_from_radians:
+ * @q: a #graphene_quaternion_t
+ * @rad_x: rotation angle on the X axis (yaw), in radians
+ * @rad_y: rotation angle on the Y axis (pitch), in radians
+ * @rad_z: rotation angle on the Z axis (roll), in radians
+ *
+ * Initializes a #graphene_quaternion_t using the values of
+ * the [Euler angles](http://en.wikipedia.org/wiki/Euler_angles)
+ * on each axis.
+ *
+ * See also: graphene_quaternion_init_from_euler()
+ *
+ * Returns: (transfer none): the initialized quaternion
+ *
+ * Since: 1.0
+ */
+graphene_quaternion_t *
+graphene_quaternion_init_from_radians (graphene_quaternion_t *q,
+                                       float                  rad_x,
+                                       float                  rad_y,
+                                       float                  rad_z)
+{
   float sin_x, sin_y, sin_z;
   float cos_x, cos_y, cos_z;
 
-  rad_x = GRAPHENE_DEG_TO_RAD (deg_x) * .5f;
-  rad_y = GRAPHENE_DEG_TO_RAD (deg_y) * .5f;
-  rad_z = GRAPHENE_DEG_TO_RAD (deg_z) * .5f;
-
-  graphene_sincos (rad_x, &sin_x, &cos_x);
-  graphene_sincos (rad_y, &sin_y, &cos_y);
-  graphene_sincos (rad_z, &sin_z, &cos_z);
+  graphene_sincos (rad_x * .5f, &sin_x, &cos_x);
+  graphene_sincos (rad_y * .5f, &sin_y, &cos_y);
+  graphene_sincos (rad_z * .5f, &sin_z, &cos_z);
 
   q->x = sin_x * cos_y * cos_z + cos_x * sin_y * sin_z;
   q->y = cos_x * sin_y * cos_z - sin_x * cos_y * sin_z;
@@ -360,7 +380,7 @@ graphene_quaternion_init_from_angles (graphene_quaternion_t *q,
  * graphene_quaternion_to_angles:
  * @q: a #graphene_quaternion_t
  * @deg_x: (out) (optional): return location for the rotation angle on
- *   the X axis (yaw), in degress
+ *   the X axis (yaw), in degrees
  * @deg_y: (out) (optional): return location for the rotation angle on
  *   the Y axis (pitch), in degrees
  * @deg_z: (out) (optional): return location for the rotation angle on
@@ -377,6 +397,40 @@ graphene_quaternion_to_angles (const graphene_quaternion_t *q,
                                float                       *deg_x,
                                float                       *deg_y,
                                float                       *deg_z)
+{
+  float rad_x, rad_y, rad_z;
+
+  graphene_quaternion_to_radians (q, &rad_x, &rad_y, &rad_z);
+
+  if (deg_x != NULL)
+    *deg_x = GRAPHENE_RAD_TO_DEG (rad_x);
+  if (deg_y != NULL)
+    *deg_y = GRAPHENE_RAD_TO_DEG (rad_y);
+  if (deg_z != NULL)
+    *deg_z = GRAPHENE_RAD_TO_DEG (rad_z);
+}
+
+/**
+ * graphene_quaternion_to_radians:
+ * @q: a #graphene_quaternion_t
+ * @rad_x: (out) (optional): return location for the rotation angle on
+ *   the X axis (yaw), in radians
+ * @rad_y: (out) (optional): return location for the rotation angle on
+ *   the Y axis (pitch), in radians
+ * @rad_z: (out) (optional): return location for the rotation angle on
+ *   the Z axis (roll), in radians
+ *
+ * Converts a #graphene_quaternion_t to its corresponding rotations
+ * on the [Euler angles](http://en.wikipedia.org/wiki/Euler_angles)
+ * on each axis.
+ *
+ * Since: 1.2
+ */
+void
+graphene_quaternion_to_radians (const graphene_quaternion_t *q,
+                                float                       *rad_x,
+                                float                       *rad_y,
+                                float                       *rad_z)
 {
   graphene_vec4_t v;
   graphene_vec4_t sq;
@@ -395,26 +449,14 @@ graphene_quaternion_to_angles (const graphene_quaternion_t *q,
   sqz = graphene_vec4_get_z (&sq);
   sqw = graphene_vec4_get_w (&sq);
 
-  if (deg_x != NULL)
-    {
-      float res = atan2f (2 * (qx * qw - qy * qz), (sqw - sqx - sqy + sqz));
+  if (rad_x != NULL)
+    *rad_x = atan2f (2 * (qx * qw - qy * qz), (sqw - sqx - sqy + sqz));
 
-      *deg_x = GRAPHENE_RAD_TO_DEG (res);
-    }
+  if (rad_y != NULL)
+    *rad_y = asinf (CLAMP (2 * ( qx * qz + qy * qw), -1, 1));
 
-  if (deg_y != NULL)
-    {
-      float res = asinf (CLAMP (2 * ( qx * qz + qy * qw), -1, 1));
-
-      *deg_y = GRAPHENE_RAD_TO_DEG (res);
-    }
-
-  if (deg_z != NULL)
-    {
-      float res = atan2f (2 * (qz * qw - qx * qy), (sqw + sqx - sqy - sqz));
-
-      *deg_z = GRAPHENE_RAD_TO_DEG (res);
-    }
+  if (rad_z != NULL)
+    *rad_z = atan2f (2 * (qz * qw - qx * qy), (sqw + sqx - sqy - sqz));
 }
 
 /**
@@ -504,69 +546,78 @@ graphene_quaternion_init_from_euler (graphene_quaternion_t  *q,
                                      const graphene_euler_t *e)
 {
   graphene_euler_order_t order = graphene_euler_get_order (e);
-  float ex = GRAPHENE_DEG_TO_RAD (graphene_euler_get_x (e)) / 2.f;
-  float ey = GRAPHENE_DEG_TO_RAD (graphene_euler_get_y (e)) / 2.f;
-  float ez = GRAPHENE_DEG_TO_RAD (graphene_euler_get_z (e)) / 2.f;
-  float x = 0.f, y = 0.f, z = 0.f, w = 1.f;
+  const float ex = GRAPHENE_DEG_TO_RAD (graphene_euler_get_x (e)) / 2.f;
+  const float ey = GRAPHENE_DEG_TO_RAD (graphene_euler_get_y (e)) / 2.f;
+  const float ez = GRAPHENE_DEG_TO_RAD (graphene_euler_get_z (e)) / 2.f;
   float c1, c2, c3, s1, s2, s3;
+  float s1c2c3, c1s2s3, c1s2c3, s1c2s3;
+  float c1c2s3, s1s2c3, c1c2c3, s1s2s3;
 
   graphene_sincos (ex, &s1, &c1);
   graphene_sincos (ey, &s2, &c2);
   graphene_sincos (ez, &s3, &c3);
 
+  s1c2c3 = s1 * c2 * c3;
+  c1s2s3 = c1 * s2 * s3;
+  c1s2c3 = c1 * s2 * c3;
+  s1c2s3 = s1 * c2 * s3;
+  c1c2s3 = c1 * c2 * s3;
+  s1s2c3 = s1 * s2 * c3;
+  c1c2c3 = c1 * c2 * c3;
+  s1s2s3 = s1 * s2 * s3;
+
   switch (order)
     {
     case GRAPHENE_EULER_ORDER_XYZ:
-      x = s1 * c2 * c3 + c1 * s2 * s3;
-      y = c1 * s2 * c3 - s1 * c2 * s3;
-      z = c1 * c2 * s3 + s1 * s2 * c3;
-      w = c1 * c2 * c3 - s1 * s2 * s3;
+      q->x = s1c2c3 + c1s2s3;
+      q->y = c1s2c3 - s1c2s3;
+      q->z = c1c2s3 + s1s2c3;
+      q->w = c1c2c3 - s1s2s3;
       break;
 
     case GRAPHENE_EULER_ORDER_YXZ:
-      x = s1 * c2 * c3 + c1 * s2 * s3;
-      y = c1 * s2 * c3 - s1 * c2 * s3;
-      z = c1 * c2 * s3 - s1 * s2 * c3;
-      w = c1 * c2 * c3 + s1 * s2 * s3;
+      q->x = s1c2c3 + c1s2s3;
+      q->y = c1s2c3 - s1c2s3;
+      q->z = c1c2s3 - s1s2c3;
+      q->w = c1c2c3 + s1s2s3;
       break;
 
     case GRAPHENE_EULER_ORDER_ZXY:
-      x = s1 * c2 * c3 - c1 * s2 * s3;
-      y = c1 * s2 * c3 + s1 * c2 * s3;
-      z = c1 * c2 * s3 + s1 * s2 * c3;
-      w = c1 * c2 * c3 - s1 * s2 * s3;
+      q->x = s1c2c3 - c1s2s3;
+      q->y = c1s2c3 + s1c2s3;
+      q->z = c1c2s3 + s1s2c3;
+      q->w = c1c2c3 - s1s2s3;
       break;
 
     case GRAPHENE_EULER_ORDER_ZYX:
-      x = s1 * c2 * c3 - c1 * s2 * s3;
-      y = c1 * s2 * c3 + s1 * c2 * s3;
-      z = c1 * c2 * s3 - s1 * s2 * c3;
-      w = c1 * c2 * c3 + s1 * s2 * s3;
+      q->x = s1c2c3 - c1s2s3;
+      q->y = c1s2c3 + s1c2s3;
+      q->z = c1c2s3 - s1s2c3;
+      q->w = c1c2c3 + s1s2s3;
       break;
 
     case GRAPHENE_EULER_ORDER_YZX:
-      x = s1 * c2 * c3 + c1 * s2 * s3;
-      y = c1 * s2 * c3 + s1 * c2 * s3;
-      z = c1 * c2 * s3 - s1 * s2 * c3;
-      w = c1 * c2 * c3 - s1 * s2 * s3;
+      q->x = s1c2c3 + c1s2s3;
+      q->y = c1s2c3 + s1c2s3;
+      q->z = c1c2s3 - s1s2c3;
+      q->w = c1c2c3 - s1s2s3;
       break;
 
     case GRAPHENE_EULER_ORDER_XZY:
-      x = s1 * c2 * c3 - c1 * s2 * s3;
-      y = c1 * s2 * c3 - s1 * c2 * s3;
-      z = c1 * c2 * s3 + s1 * s2 * c3;
-      w = c1 * c2 * c3 + s1 * s2 * s3;
+      q->x = s1c2c3 - c1s2s3;
+      q->y = c1s2c3 - s1c2s3;
+      q->z = c1c2s3 + s1s2c3;
+      q->w = c1c2c3 + s1s2s3;
       break;
 
     case GRAPHENE_EULER_ORDER_DEFAULT:
+      q->x = 0.f;
+      q->y = 0.f;
+      q->z = 0.f;
+      q->w = 1.f;
       break;
     }
 
-  q->x = x;
-  q->y = y;
-  q->z = z;
-  q->w = w;
-
   return q;
 }
 
