@@ -29,12 +29,15 @@
  * #graphene_rect_t is a type representing a rectangle through an origin
  * #graphene_point_t point and a #graphene_size_t size.
  *
+ * ![](rectangle.png)
+ *
  * Operations on a #graphene_rect_t will normalize the rectangle, to
  * ensure that the origin is always the top-left corner of the rectangle
  * and that the size is always positive.
  */
 
 #include "graphene-private.h"
+
 #include "graphene-rect.h"
 
 #include <math.h>
@@ -45,12 +48,9 @@
  *
  * Normalizes the passed #graphene_rect_t.
  */
-static void
+static inline void
 graphene_rect_normalize_in_place (graphene_rect_t *r)
 {
-  if (likely (r->size.width >= 0.f && r->size.height >= 0.f))
-    return;
-
   if (r->size.width < 0.f)
     {
       float size = fabsf (r->size.width);
@@ -211,6 +211,30 @@ graphene_rect_normalize (graphene_rect_t *r)
   return r;
 }
 
+/**
+ * graphene_rect_normalize_r:
+ * @r: a #graphene_rect_t
+ * @res: (out caller-allocates): the return location for the
+ *   normalized rectangle
+ *
+ * Normalizes the passed rectangle.
+ *
+ * This function ensures that the size of the rectangle is made of
+ * positive values, and that the origin is in the top-left corner
+ * of the rectangle.
+ *
+ * Since: 1.4
+ */
+void
+graphene_rect_normalize_r (const graphene_rect_t *r,
+                           graphene_rect_t       *res)
+{
+  if (res != r)
+    *res = *r;
+
+  graphene_rect_normalize_in_place (res);
+}
+
 /**
  * graphene_rect_get_center:
  * @r: a #graphene_rect_t
@@ -320,6 +344,30 @@ graphene_rect_get_bottom_right (const graphene_rect_t  *r,
                        rr.origin.y + rr.size.height);
 }
 
+/**
+ * graphene_rect_get_vertices:
+ * @r: a #graphene_rect_t
+ * @vertices: (out) (array fixed-size=4): return location for an array
+ *  of 4 #graphene_vec2_t
+ *
+ * Computes the four vertices of a #graphene_rect_t.
+ *
+ * Since: 1.4
+ */
+void
+graphene_rect_get_vertices (const graphene_rect_t *r,
+                            graphene_vec2_t        vertices[])
+{
+  graphene_rect_t rr;
+
+  graphene_rect_normalize_r (r, &rr);
+
+  graphene_vec2_init (&vertices[0], rr.origin.x, rr.origin.y);
+  graphene_vec2_init (&vertices[1], rr.origin.x + rr.size.width, rr.origin.y);
+  graphene_vec2_init (&vertices[2], rr.origin.x + rr.size.width, rr.origin.y + rr.size.height);
+  graphene_vec2_init (&vertices[3], rr.origin.x, rr.origin.y + rr.size.height);
+}
+
 #define GRAPHENE_RECT_GET(arg,part,field) \
 float \
 graphene_rect_get_##field (const graphene_rect_t *arg) \
@@ -392,6 +440,10 @@ GRAPHENE_RECT_GET (r, size, height)
  *
  * Computes the union of the two given rectangles.
  *
+ * ![](rectangle-union.png)
+ *
+ * The union in the image above is the blue outline.
+ *
  * Since: 1.0
  */
 void
@@ -410,8 +462,8 @@ graphene_rect_union (const graphene_rect_t *a,
   res->origin.x = MIN (ra.origin.x, rb.origin.x);
   res->origin.y = MIN (ra.origin.y, rb.origin.y);
 
-  res->size.width = MAX (ra.size.width, rb.size.width);
-  res->size.height = MAX (ra.size.height, rb.size.height);
+  res->size.width = MAX (ra.origin.x + ra.size.width, rb.origin.x + rb.size.width) - res->origin.x;
+  res->size.height = MAX (ra.origin.y + ra.size.height, rb.origin.y + rb.size.height) - res->origin.y;
 }
 
 /**
@@ -423,6 +475,10 @@ graphene_rect_union (const graphene_rect_t *a,
  *
  * Computes the intersection of the two given rectangles.
  *
+ * ![](rectangle-intersection.png)
+ *
+ * The intersection in the image above is the blue outline.
+ *
  * If the two rectangles do not intersect, @res will contain
  * a degenerate rectangle with origin in (0, 0) and a size of 0.
  *
@@ -531,14 +587,37 @@ graphene_rect_offset (graphene_rect_t *r,
                       float            d_x,
                       float            d_y)
 {
-  graphene_rect_normalize_in_place (r);
-
-  r->origin.x += d_x;
-  r->origin.y += d_y;
+  graphene_rect_offset_r (r, d_x, d_y, r);
 
   return r;
 }
 
+/**
+ * graphene_rect_offset_r:
+ * @r: a #graphene_rect_t
+ * @d_x: the horizontal offset
+ * @d_y: the vertical offset
+ * @res: (out caller-allocates): return location for the offset
+ *   rectangle
+ *
+ * Offsets the origin of the given rectangle by @d_x and @d_y.
+ *
+ * The size of the rectangle is left unchanged.
+ *
+ * Since: 1.4
+ */
+void
+graphene_rect_offset_r (const graphene_rect_t *r,
+                        float                  d_x,
+                        float                  d_y,
+                        graphene_rect_t       *res)
+{
+  graphene_rect_normalize_r (r, res);
+
+  res->origin.x += d_x;
+  res->origin.y += d_y;
+}
+
 /**
  * graphene_rect_inset:
  * @r: a #graphene_rect_t
@@ -553,7 +632,7 @@ graphene_rect_offset (graphene_rect_t *r,
  * values.
  *
  * The origin of the rectangle is offset by @d_x and @d_y, while the size
- * is adjusted by (2 * @d_x, 2 * @d_y). If @d_x and @d_y are positive
+ * is adjusted by `(2 * @d_x, 2 * @d_y)`. If @d_x and @d_y are positive
  * values, the size of the rectangle is decreased; if @d_x and @d_y are
  * negative values, the size of the rectangle is increased.
  *
@@ -569,30 +648,63 @@ graphene_rect_inset (graphene_rect_t *r,
                      float            d_x,
                      float            d_y)
 {
-  graphene_rect_normalize_in_place (r);
-
-  r->origin.x += d_x;
-  r->origin.y += d_y;
-
-  if (d_x >= 0.f)
-    r->size.width -= (d_x * 2.f);
-  else
-    r->size.width += (d_x * -2.f);
-
-  if (d_y >= 0.f)
-    r->size.height -= (d_y * 2.f);
-  else
-    r->size.height += (d_y * -2.f);
-
-  if (r->size.width < 0.f)
-    r->size.width = 0.f;
-
-  if (r->size.height < 0.f)
-    r->size.height = 0.f;
+  graphene_rect_inset_r (r, d_x, d_y, r);
 
   return r;
 }
 
+/**
+ * graphene_rect_inset_r:
+ * @r: a #graphene_rect_t
+ * @d_x: the horizontal inset
+ * @d_y: the vertical inset
+ * @res: (out caller-allocates): return location for the inset rectangle
+ *
+ * Changes the given rectangle to be smaller, or larger depending on the
+ * given inset parameters.
+ *
+ * To create an inset rectangle, use positive @d_x or @d_y values; to
+ * create a larger, encompassing rectangle, use negative @d_x or @d_y
+ * values.
+ *
+ * The origin of the rectangle is offset by @d_x and @d_y, while the size
+ * is adjusted by `(2 * @d_x, 2 * @d_y)`. If @d_x and @d_y are positive
+ * values, the size of the rectangle is decreased; if @d_x and @d_y are
+ * negative values, the size of the rectangle is increased.
+ *
+ * If the size of the resulting inset rectangle has a negative width or
+ * height then the size will be set to zero.
+ *
+ * Since: 1.4
+ */
+void
+graphene_rect_inset_r (const graphene_rect_t *r,
+                       float                  d_x,
+                       float                  d_y,
+                       graphene_rect_t       *res)
+{
+  graphene_rect_normalize_r (r, res);
+
+  res->origin.x += d_x;
+  res->origin.y += d_y;
+
+  if (d_x >= 0.f)
+    res->size.width -= (d_x * 2.f);
+  else
+    res->size.width += (d_x * -2.f);
+
+  if (d_y >= 0.f)
+    res->size.height -= (d_y * 2.f);
+  else
+    res->size.height += (d_y * -2.f);
+
+  if (res->size.width < 0.f)
+    res->size.width = 0.f;
+
+  if (res->size.height < 0.f)
+    res->size.height = 0.f;
+}
+
 /**
  * graphene_rect_round_to_pixel:
  * @r: a #graphene_rect_t
@@ -604,21 +716,68 @@ graphene_rect_inset (graphene_rect_t *r,
  * Returns: (transfer none): the pixel-aligned rectangle.
  *
  * Since: 1.0
+ *
+ * Deprecated: 1.4: Use graphene_rect_round() instead
  */
 graphene_rect_t *
 graphene_rect_round_to_pixel (graphene_rect_t *r)
 {
-  graphene_rect_normalize_in_place (r);
-
-  r->origin.x = floorf (r->origin.x);
-  r->origin.y = floorf (r->origin.y);
-
-  r->size.width = ceilf (r->size.width);
-  r->size.height = ceilf (r->size.height);
+  graphene_rect_round (r, r);
 
   return r;
 }
 
+/**
+ * graphene_rect_round:
+ * @r: a #graphene_rect_t
+ * @res: (out caller-allocates): return location for the
+ *   rounded rectangle
+ *
+ * Rounds the origin and size of the given rectangle to
+ * their nearest integer values; the rounding is guaranteed
+ * to be large enough to contain the original rectangle.
+ *
+ * This function is the equivalent of calling `floor` on
+ * the coordinates of the origin, and `ceil` on the size.
+ *
+ * Since: 1.4
+ */
+void
+graphene_rect_round (const graphene_rect_t *r,
+                     graphene_rect_t       *res)
+{
+  graphene_rect_normalize_r (r, res);
+
+  res->origin.x = floorf (res->origin.x);
+  res->origin.y = floorf (res->origin.y);
+
+  res->size.width = ceilf (res->size.width);
+  res->size.height = ceilf (res->size.height);
+}
+
+/**
+ * graphene_rect_expand:
+ * @r: a #graphene_rect_t
+ * @p: a #graphene_point_t
+ * @res: (out caller-allocates): return location for the expanded rectangle
+ *
+ * Expands a #graphene_rect_t to contain the given #graphene_point_t.
+ *
+ * Since: 1.4
+ */
+void
+graphene_rect_expand (const graphene_rect_t  *r,
+                      const graphene_point_t *p,
+                      graphene_rect_t        *res)
+{
+  graphene_rect_t tmp;
+
+  graphene_rect_init (&tmp, p->x, p->y, 0.f, 0.f);
+  graphene_rect_union (r, &tmp, res);
+
+  graphene_rect_normalize_in_place (res);
+}
+
 /**
  * graphene_rect_interpolate:
  * @a: a #graphene_rect_t
@@ -651,3 +810,21 @@ graphene_rect_interpolate (const graphene_rect_t *a,
   res->size.width = graphene_lerp (ra.size.width, rb.size.width, factor);
   res->size.height = graphene_lerp (ra.size.height, rb.size.height, factor);
 }
+
+static const graphene_rect_t _graphene_rect_zero = GRAPHENE_RECT_INIT (0, 0, 0, 0);
+
+/**
+ * graphene_rect_zero:
+ *
+ * Returns a degenerate rectangle with origin fixed at (0, 0) and
+ * a size of 0, 0.
+ *
+ * Returns: (transfer none): a fixed rectangle
+ *
+ * Since: 1.4
+ */
+const graphene_rect_t *
+graphene_rect_zero (void)
+{
+  return &_graphene_rect_zero;
+}
