@@ -288,7 +288,16 @@ float
  * @v: a floating point value
  *
  * Sets all the components of a new #graphene_simd4f_t to the
- * same value @v.
+ * same value @v:
+ *
+ * |[<!-- language="plain" -->
+ *   {
+ *     .x = v,
+ *     .y = v,
+ *     .z = v,
+ *     .w = v
+ *   };
+ * ]|
  *
  * Returns: the initialized #graphene_simd4f_t
  *
@@ -305,7 +314,16 @@ graphene_simd4f_t
  * @s: a #graphene_simd4f_t
  *
  * Sets all the components of a new #graphene_simd4f_t to the
- * same value of the first component of the passed vector.
+ * same value of the first component of the passed vector:
+ *
+ * |[<!-- language="plain" -->
+ *   {
+ *     .x = s.x,
+ *     .y = s.x,
+ *     .z = s.x,
+ *     .w = s.x
+ *   }
+ * ]|
  *
  * Returns: the initialized #graphene_simd4f_t
  *
@@ -322,7 +340,16 @@ graphene_simd4f_t
  * @s: a #graphene_simd4f_t
  *
  * Sets all the components of a new #graphene_simd4f_t to the
- * same value of the second component of the passed vector.
+ * same value of the second component of the passed vector:
+ *
+ * |[<!-- language="plain" -->
+ *   {
+ *     .x = s.y,
+ *     .y = s.y,
+ *     .z = s.y,
+ *     .w = s.y
+ *   }
+ * ]|
  *
  * Returns: the initialized #graphene_simd4f_t
  *
@@ -339,7 +366,16 @@ graphene_simd4f_t
  * @s: a #graphene_simd4f_t
  *
  * Sets all the components of a #graphene_simd4f_t to the
- * same value of the third component of the passed vector.
+ * same value of the third component of the passed vector:
+ *
+ * |[<!-- language="plain" -->
+ *   {
+ *     .x = s.z,
+ *     .y = s.z,
+ *     .z = s.z,
+ *     .w = s.z
+ *   }
+ * ]|
  *
  * Returns: the initialized #graphene_simd4f_t
  *
@@ -356,7 +392,16 @@ graphene_simd4f_t
  * @s: a #graphene_simd4f_t
  *
  * Sets all the components of a #graphene_simd4f_t to the
- * same value of the fourth component of the passed vector.
+ * same value of the fourth component of the passed vector:
+ *
+ * |[<!-- language="plain" -->
+ *   {
+ *     .x = s.w,
+ *     .y = s.w,
+ *     .z = s.w,
+ *     .w = s.w
+ *   }
+ * ]|
  *
  * Returns: the initialized #graphene_simd4f_t
  *
@@ -374,6 +419,15 @@ graphene_simd4f_t
  *
  * Computes the reciprocal of every component of @s.
  *
+ * |[<!-- language="plain" -->
+ *   {
+ *     .x = 1.0 / s.x,
+ *     .y = 1.0 / s.y,
+ *     .z = 1.0 / s.z,
+ *     .w = 1.0 / s.w
+ *   }
+ * ]|
+ *
  * Returns: a vector containing the reciprocal of the
  *   passed vector
  *
@@ -391,7 +445,7 @@ graphene_simd4f_t
  *
  * Computes the square root of every component of @s.
  *
- * |[<!-- language="C" -->
+ * |[<!-- language="plain" -->
  *   {
  *     .x = sqrt (s.x),
  *     .y = sqrt (s.y),
@@ -418,7 +472,7 @@ graphene_simd4f_t
  * Computes the reciprocal square root of every component
  * of @s.
  *
- * |[<!-- language="C" -->
+ * |[<!-- language="plain" -->
  *   {
  *     .x = 1.0 / sqrt (s.x),
  *     .y = 1.0 / sqrt (s.y),
@@ -447,7 +501,7 @@ graphene_simd4f_t
  * component is the sum of the respective components
  * in @a and @b.
  *
- * |[<!-- lanugage="C" -->
+ * |[<!-- lanugage="plain" -->
  *   {
  *     .x = a.x + b.x,
  *     .y = a.y + b.y,
@@ -476,7 +530,7 @@ graphene_simd4f_t
  * component is the subtraction of the respective components
  * in @a and @b.
  *
- * |[<!-- lanugage="C" -->
+ * |[<!-- lanugage="plain" -->
  *   {
  *     .x = a.x - b.x,
  *     .y = a.y - b.y,
@@ -505,7 +559,7 @@ graphene_simd4f_t
  * component is the multiplication of the respective components
  * in @a and @b.
  *
- * |[<!-- lanugage="C" -->
+ * |[<!-- lanugage="plain" -->
  *   {
  *     .x = a.x * b.x,
  *     .y = a.y * b.y,
@@ -534,7 +588,7 @@ graphene_simd4f_t
  * component is the division of the respective components
  * in @a and @b.
  *
- * |[<!-- lanugage="C" -->
+ * |[<!-- lanugage="plain" -->
  *   {
  *     .x = a.x / b.x,
  *     .y = a.y / b.y,
@@ -561,8 +615,7 @@ graphene_simd4f_t
  *
  * Creates a new #graphene_simd4f_t vector where each
  * component contains the 3-way cross product of the
- * given @a and @b vectors. You can access the result
- * as a floating point value through graphene_simd4f_get_x().
+ * given @a and @b vectors.
  *
  * Returns: the cross3 vector
  *
@@ -575,6 +628,45 @@ graphene_simd4f_t
   return graphene_simd4f_cross3 (a, b);
 }
 
+/**
+ * graphene_simd4f_dot3:
+ * @a: a #graphene_simd4f_t
+ * @b: a #graphene_simd4f_t
+ *
+ * Computes the dot product of the first three components of the
+ * two given #graphene_simd4f_t.
+ *
+ * Returns: a vector whose components are all set to the
+ *   dot product of the components of the two operands
+ *
+ * Since: 1.0
+ */
+graphene_simd4f_t
+(graphene_simd4f_dot3) (const graphene_simd4f_t a,
+                        const graphene_simd4f_t b)
+{
+  return graphene_simd4f_dot3 (a, b);
+}
+
+/**
+ * graphene_simd4f_dot3_scalar:
+ * @a: a #graphene_simd4f_t
+ * @b: a #graphene_simd4f_t
+ *
+ * Computes the dot product of the first three components of the
+ * two given #graphene_simd4f_t.
+ *
+ * Returns: the dot product of the two vectors, as a scalar value.
+ *
+ * Since: 1.4
+ */
+float
+(graphene_simd4f_dot3_scalar) (const graphene_simd4f_t a,
+                               const graphene_simd4f_t b)
+{
+  return graphene_simd4f_dot3_scalar (a, b);
+}
+
 /**
  * graphene_simd4f_min:
  * @a: a #graphene_simd4f_t
@@ -1068,10 +1160,10 @@ graphene_simd4f_t
 (graphene_simd4f_reciprocal) (graphene_simd4f_t v)
 {
   graphene_simd4f_t s = {
-    1.0f / v.x,
-    1.0f / v.y,
-    1.0f / v.z,
-    1.0f / v.w
+    v.x != 0.f ? 1.0f / v.x : 0.f,
+    v.y != 0.f ? 1.0f / v.y : 0.f,
+    v.z != 0.f ? 1.0f / v.z : 0.f,
+    v.w != 0.f ? 1.0f / v.w : 0.f
   };
   return s;
 }
@@ -1080,10 +1172,10 @@ graphene_simd4f_t
 (graphene_simd4f_sqrt) (graphene_simd4f_t v)
 {
   graphene_simd4f_t s = {
-    sqrt (v.x),
-    sqrt (v.y),
-    sqrt (v.z),
-    sqrt (v.w)
+    sqrtf (v.x),
+    sqrtf (v.y),
+    sqrtf (v.z),
+    sqrtf (v.w)
   };
   return s;
 }
@@ -1092,10 +1184,10 @@ graphene_simd4f_t
 (graphene_simd4f_rsqrt) (graphene_simd4f_t v)
 {
   graphene_simd4f_t s = {
-    1.0f / sqrt (v.x),
-    1.0f / sqrt (v.y),
-    1.0f / sqrt (v.z),
-    1.0f / sqrt (v.w)
+    v.x != 0.f ? 1.0f / sqrtf (v.x) : 0.f,
+    v.y != 0.f ? 1.0f / sqrtf (v.y) : 0.f,
+    v.z != 0.f ? 1.0f / sqrtf (v.z) : 0.f,
+    v.w != 0.f ? 1.0f / sqrtf (v.w) : 0.f
   };
   return s;
 }
@@ -1144,10 +1236,10 @@ graphene_simd4f_t
                        const graphene_simd4f_t b)
 {
   graphene_simd4f_t s = {
-    a.x / b.x,
-    a.y / b.y,
-    a.z / b.z,
-    a.w / b.w
+    b.x != 0.f ? a.x / b.x : 0.f,
+    b.y != 0.f ? a.y / b.y : 0.f,
+    b.z != 0.f ? a.z / b.z : 0.f,
+    b.w != 0.f ? a.w / b.w : 0.f
   };
   return s;
 }
@@ -1162,6 +1254,20 @@ graphene_simd4f_t
                                0.f);
 }
 
+graphene_simd4f_t
+(graphene_simd4f_dot3) (const graphene_simd4f_t a,
+                        const graphene_simd4f_t b)
+{
+  return graphene_simd4f_splat (graphene_simd4f_dot3_scalar (a, b));
+}
+
+float
+(graphene_simd4f_dot3_scalar) (const graphene_simd4f_t a,
+                               const graphene_simd4f_t b)
+{
+  return a.x * b.x + a.y * b.y + a.z * b.z;
+}
+
 graphene_simd4f_t
 (graphene_simd4f_min) (const graphene_simd4f_t a,
                        const graphene_simd4f_t b)
