@@ -27,6 +27,21 @@ GRAPHENE_TEST_UNIT_BEGIN (matrix_identity)
 
   graphene_matrix_scale (&m, 2.0f, 2.0f, 2.0f);
   g_assert_false (graphene_matrix_is_identity (&m));
+
+  v[0]  = 1.f; v[4]  = 0.f; v[8]  = 0.f; v[12] = 0.f;
+  v[1]  = 0.f; v[5]  = 1.f; v[9]  = 0.f; v[13] = 0.f;
+  v[2]  = 0.f; v[6]  = 0.f; v[10] = 1.f; v[14] = 0.f;
+  v[3]  = 0.f; v[7]  = 0.f; v[11] = 0.f; v[15] = 1.f;
+
+  graphene_matrix_init_from_float (&m, v);
+  g_assert_true (graphene_matrix_is_identity (&m));
+
+  graphene_matrix_init_from_vec4 (&m,
+                                  graphene_vec4_x_axis (),
+                                  graphene_vec4_y_axis (),
+                                  graphene_vec4_z_axis (),
+                                  graphene_vec4_w_axis ());
+  g_assert_true (graphene_matrix_is_identity (&m));
 }
 GRAPHENE_TEST_UNIT_END
 
@@ -58,17 +73,6 @@ GRAPHENE_TEST_UNIT_END
 
 GRAPHENE_TEST_UNIT_BEGIN (matrix_rotation_euler_quaternion)
 {
-/* XXX: disable on GCC < 4.9; it seems the compiler emits wrong floating
- *      point instructions with -ffast-math
- */
-#if !defined(__clang__) && defined(__GNUC__) && __GNUC__ < 4 || (__GNUC__ == 4 && __GNUC_MINOR__ < 9)
-# if GLIB_CHECK_VERSION (2, 38, 0)
-  g_test_skip ("Disabled on GCC < 4.9");
-# else
-  if (g_test_verbose ())
-    g_test_message ("Disabled on GCC < 4.9");
-# endif
-#else
   graphene_matrix_t m0, m1, m2;
   graphene_quaternion_t q;
   graphene_euler_t e;
@@ -87,7 +91,6 @@ GRAPHENE_TEST_UNIT_BEGIN (matrix_rotation_euler_quaternion)
   graphene_assert_fuzzy_matrix_equal (&m0, &m1, 0.0001f);
   graphene_assert_fuzzy_matrix_equal (&m0, &m2, 0.0001f);
   graphene_assert_fuzzy_matrix_equal (&m1, &m2, 0.0001f);
-#endif
 }
 GRAPHENE_TEST_UNIT_END
 
@@ -168,8 +171,7 @@ GRAPHENE_TEST_UNIT_BEGIN (matrix_invert)
   graphene_matrix_multiply (&m, &inv, &res);
   graphene_assert_fuzzy_matrix_equal (&identity, &res, 0.0001f);
 
-  graphene_matrix_translate (&m,
-			     graphene_point3d_init (&p, 1,2,3));
+  graphene_matrix_translate (&m, graphene_point3d_init (&p, 1.f, 2.f, 3.f));
 
   graphene_matrix_inverse (&m , &inv);
   graphene_matrix_multiply (&m, &inv, &res);
@@ -432,6 +434,52 @@ GRAPHENE_TEST_UNIT_BEGIN (matrix_2d_round_trip)
 }
 GRAPHENE_TEST_UNIT_END
 
+GRAPHENE_TEST_UNIT_BEGIN (matrix_interpolate)
+{
+  graphene_matrix_t m1, m2, m3, mr;
+
+  graphene_matrix_init_identity (&m1);
+  graphene_matrix_init_identity (&m2);
+
+  graphene_matrix_translate (&m2, &(graphene_point3d_t) { 100.f, 100.f, 100.f });
+  g_assert_true (!graphene_matrix_is_2d (&m2));
+
+  graphene_matrix_interpolate (&m1, &m2, 0.0, &mr);
+  graphene_assert_fuzzy_matrix_equal (&mr, &m1, 0.1);
+
+  graphene_matrix_interpolate (&m1, &m2, 1.0, &mr);
+  graphene_assert_fuzzy_matrix_equal (&mr, &m2, 0.1);
+
+  graphene_matrix_init_translate (&m3, &(graphene_point3d_t) { 50.f, 50.f, 50.f });
+  graphene_matrix_interpolate (&m1, &m2, 0.5, &mr);
+  graphene_assert_fuzzy_matrix_equal (&mr, &m3, 0.1);
+}
+GRAPHENE_TEST_UNIT_END
+
+GRAPHENE_TEST_UNIT_BEGIN (matrix_2d_interpolate)
+{
+  graphene_matrix_t m1, m2, m3, mr;
+
+  graphene_matrix_init_identity (&m1);
+  graphene_matrix_init_identity (&m2);
+
+  graphene_matrix_translate (&m2, &(graphene_point3d_t) { 100.f, 100.f, 0.f });
+
+  g_assert_true (graphene_matrix_is_2d (&m1));
+  g_assert_true (graphene_matrix_is_2d (&m2));
+
+  graphene_matrix_interpolate (&m1, &m2, 0.0, &mr);
+  graphene_assert_fuzzy_matrix_equal (&mr, &m1, 0.01);
+
+  graphene_matrix_interpolate (&m1, &m2, 1.0, &mr);
+  graphene_assert_fuzzy_matrix_equal (&mr, &m2, 0.01);
+
+  graphene_matrix_init_translate (&m3, &(graphene_point3d_t) { 50.f, 50.f, 0.f });
+  graphene_matrix_interpolate (&m1, &m2, 0.5, &mr);
+  graphene_assert_fuzzy_matrix_equal (&mr, &m3, 0.01);
+}
+GRAPHENE_TEST_UNIT_END
+
 GRAPHENE_TEST_SUITE (
   GRAPHENE_TEST_UNIT ("/matrix/identity", matrix_identity)
   GRAPHENE_TEST_UNIT ("/matrix/scale", matrix_scale)
@@ -441,7 +489,9 @@ GRAPHENE_TEST_SUITE (
   GRAPHENE_TEST_UNIT ("/matrix/neutral_element", matrix_neutral_element)
   GRAPHENE_TEST_UNIT ("/matrix/look_at", matrix_look_at)
   GRAPHENE_TEST_UNIT ("/matrix/invert", matrix_invert)
+  GRAPHENE_TEST_UNIT ("/matrix/interpolate", matrix_interpolate)
   GRAPHENE_TEST_UNIT ("/matrix/2d/identity", matrix_2d_identity)
   GRAPHENE_TEST_UNIT ("/matrix/2d/transforms", matrix_2d_transforms)
   GRAPHENE_TEST_UNIT ("/matrix/2d/round-trip", matrix_2d_round_trip)
+  GRAPHENE_TEST_UNIT ("/matrix/2d/interpolate", matrix_2d_interpolate)
 )
